# Copyright 2024 Canonical Ltd.
# See LICENSE file for licensing details.

"""Utils for observability Juju charms."""

import importlib

__all__ = [
    "Coordinator",
    "ClusterProvider",
    "ClusterRequirer",
    "Nginx",
    "NginxPrometheusExporter",
    "Worker",
]

current_package = __package__


class _LazyModule:
    def __init__(self, module_name: str):
        self.module_name = module_name
        self.module = None

    def _load(self):
        if self.module is None:
            self.module = importlib.import_module(self.module_name, current_package)
        return self.module

    def __getattr__(self, item: str):
        module = self._load()
        return getattr(module, item)


# Create lazy-loaded modules
Coordinator = _LazyModule(".coordinator")
ClusterProvider = _LazyModule(".interface")
ClusterRequirer = _LazyModule(".interface")
Nginx = _LazyModule(".nginx")
NginxPrometheusExporter = _LazyModule(".nginx")
Worker = _LazyModule(".worker")



# Copyright 2024 Canonical Ltd.
# See LICENSE file for licensing details.

import unittest

from cosl.mandatory_relation_pairs import MandatoryRelationPairs


class TestMandatoryRelationPairs(unittest.TestCase):
    def setUp(self):
        self.rels = {
            "juju-info": [  # must be paired with:
                {"grafana-cloud-config"},  # or
                {"send-remote-write", "logging-consumer"},
            ],
            "cos-agent": [  # must be paired with:
                {"grafana-cloud-config"},  # or
                {"send-remote-write", "logging-consumer", "grafana-dashboards-provider"},
            ],
        }

    def test_unspecified(self):
        mrp = MandatoryRelationPairs(self.rels.copy())
        self.assertEqual(mrp.get_missing(), {})
        self.assertEqual(mrp.get_missing("some_unspecified_relation"), {})

        # Should be ok if "outgoing" are specified without "incoming".
        self.assertEqual(mrp.get_missing("grafana-cloud-config"), {})
        self.assertEqual(mrp.get_missing("logging-consumer"), {})

    def test_incoming(self):
        """One or more of the 'incoming' relations are the only ones present."""
        mrp = MandatoryRelationPairs(self.rels.copy())
        self.assertEqual(
            mrp.get_missing("cos-agent"),
            {"cos-agent": self.rels["cos-agent"]},
        )
        self.assertEqual(mrp.get_missing(*self.rels.keys()), self.rels)

    def test_match(self):
        """An 'incoming' relation has all its required matching 'outgoing' relations."""
        mrp = MandatoryRelationPairs(self.rels.copy())
        self.assertEqual(mrp.get_missing("cos-agent", "grafana-cloud-config"), {})
        self.assertEqual(
            mrp.get_missing(
                "cos-agent",
                "grafana-cloud-config",
                "send-remote-write",
                "some-unspecified-relation",
            ),
            {},
        )
        self.assertEqual(
            mrp.get_missing(
                "cos-agent",
                "send-remote-write",
                "logging-consumer",
                "grafana-dashboards-provider",
            ),
            {},
        )
        self.assertEqual(
            mrp.get_missing(
                "cos-agent",
                "juju-info",
                "send-remote-write",
                "logging-consumer",
                "grafana-dashboards-provider",
            ),
            {},
        )

    def test_under_match(self):
        mrp = MandatoryRelationPairs(self.rels.copy())
        self.assertEqual(
            mrp.get_missing("juju-info", "send-remote-write"),
            {"juju-info": [{"grafana-cloud-config"}, {"logging-consumer"}]},
        )
        self.assertEqual(
            mrp.get_missing("cos-agent", "juju-info", "send-remote-write", "logging-consumer"),
            {"cos-agent": [{"grafana-cloud-config"}, {"grafana-dashboards-provider"}]},
        )

    def test_under_match_as_str(self):
        mrp = MandatoryRelationPairs(self.rels.copy())
        self.assertEqual(
            mrp.get_missing_as_str("juju-info"),
            "['grafana-cloud-config']|['logging-consumer', 'send-remote-write'] for juju-info",
        )
        self.assertEqual(
            mrp.get_missing_as_str("juju-info", "cos-agent"),
            "['grafana-cloud-config']|['grafana-dashboards-provider', 'logging-consumer', 'send-remote-write'] for cos-agent; ['grafana-cloud-config']|['logging-consumer', 'send-remote-write'] for juju-info",
        )



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

"""COS Tool."""

import logging
import platform
import re
import subprocess
import tempfile
import uuid
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import yaml
from typing_extensions import Concatenate, ParamSpec, TypeVar

from .types import OfficialRuleFileFormat, QueryType

logger = logging.getLogger(__name__)

_T = TypeVar("_T")
_P = ParamSpec("_P")


def ensure_querytype(func: Callable[_P, _T]) -> Callable[Concatenate["CosTool", _P], _T]:
    """A small decorator to ensure that query type is specified."""

    def wrapper(self: "CosTool", *args: _P.args, **kwargs: _P.kwargs) -> _T:
        if not self.query_type and not kwargs.get("query_type", None):
            raise TypeError(
                "Either a default query type or a per-method query type must be used for `CosTool`!"
            )
        return func(self, *args, **kwargs)  # type: ignore

    wrapper.__doc__ = func.__doc__
    return wrapper


class CosTool:
    """Uses cos-tool to inject label matchers into alert rule expressions and validate rules.

    Args:
        default_query_type: an optional querytype to use for all invocations of this class, if
          not specified per-method. Either :default_query_type: or per-method :query_type:
          **must** be used, or a :TypeError: will be raised.
    """

    _path = None
    _disabled = False
    query_type: Union[QueryType, None] = None

    def __init__(self, default_query_type: Optional[QueryType] = None):
        self.query_type = default_query_type

    @property
    def path(self):
        """Lazy lookup of the path of cos-tool."""
        if self._disabled:
            return None
        if not self._path:
            self._path = self._get_tool_path()
            if not self._path:
                logger.debug("Skipping injection of juju topology as label matchers")
                self._disabled = True
        return self._path

    @ensure_querytype
    def apply_label_matchers(
        self, rules: OfficialRuleFileFormat, query_type: Optional[QueryType] = None
    ) -> OfficialRuleFileFormat:
        """Will apply label matchers to the expression of all alerts in all supplied groups."""
        query_type = query_type or self.query_type
        if not self.path:
            return rules
        for group in rules["groups"]:
            rules_in_group = group.get("rules", [])
            for rule in rules_in_group:
                topology = {}
                # if the user for some reason has provided juju_unit, we'll need to honor it
                # in most cases, however, this will be empty
                for label in [
                    "juju_model",
                    "juju_model_uuid",
                    "juju_application",
                    "juju_charm",
                    "juju_unit",
                ]:
                    if label in rule["labels"]:
                        topology[label] = rule["labels"][label]

                rule["expr"] = self.inject_label_matchers(rule["expr"], topology, query_type)  # type: ignore
        return rules

    @ensure_querytype
    def validate_alert_rules(
        self, rules: Dict[str, Any], query_type: Optional[QueryType] = None
    ) -> Tuple[bool, str]:
        """Will validate correctness of alert rules, returning a boolean and any errors."""
        query_type = query_type or self.query_type
        if not self.path:
            logger.debug("`cos-tool` unavailable. Not validating alert correctness.")
            return True, ""

        with tempfile.TemporaryDirectory() as tmpdir:
            rule_path = Path(tmpdir + "/validate_rule.yaml")

            # Smash "our" rules format into what upstream actually uses for Loki,
            # which is more like:
            #
            # groups:
            #   - name: foo
            #     rules:
            #       - alert: SomeAlert
            #         expr: up
            #       - alert: OtherAlert
            #         expr: up
            if query_type == "logql":
                transformed_rules = {"groups": []}  # type: Dict[str, Any]
                for rule in rules["groups"]:
                    transformed = {"name": str(uuid.uuid4()), "rules": [rule]}
                    transformed_rules["groups"].append(transformed)

                rules = transformed_rules

            rule_path.write_text(yaml.dump(rules))

            args = [str(self.path), "--format", query_type, "validate", str(rule_path)]
            # noinspection PyBroadException
            try:
                self._exec(args)  # type: ignore
                return True, ""
            except subprocess.CalledProcessError as e:
                logger.debug("Validating the rules failed: %s", e.output.decode("utf-8"))
                return False, ", ".join(
                    [
                        line
                        for line in e.output.decode("utf-8").splitlines()
                        if "error validating" in line
                    ]
                )

    @ensure_querytype
    def inject_label_matchers(
        self,
        expression: str,
        topology: Dict[str, str],
        query_type: Optional[QueryType] = None,
        dashboard_variable: Optional[bool] = False,
    ) -> str:
        """Add label matchers to an expression."""
        query_type = query_type or self.query_type

        if not topology:
            return expression
        if not self.path:
            logger.debug("`cos-tool` unavailable. Leaving expression unchanged: %s", expression)
            return expression
        args = [str(self.path), "--format", query_type, "transform"]

        value_tmpl = r"${}" if dashboard_variable else "{}"

        variable_topology = {k: value_tmpl.format(topology[k]) for k in topology.keys()}
        args.extend(
            [
                "--label-matcher={}={}".format(key, value)
                for key, value in variable_topology.items()
            ]
        )

        # Pass a leading "--" so expressions with a negation or subtraction aren't interpreted as
        # flags
        args.extend(["--", "{}".format(expression)])
        # noinspection PyBroadException
        try:
            return (
                re.sub(r'="\$juju', r'=~"$juju', self._exec(args))  # type: ignore
                if dashboard_variable
                else self._exec(args)  # type: ignore
            )
        except subprocess.CalledProcessError as e:
            logger.debug('Applying the expression failed: "%s", falling back to the original', e)
            return expression

    def _get_tool_path(self) -> Optional[Path]:
        arch = platform.machine()
        arch = "amd64" if arch == "x86_64" else arch
        res = "cos-tool-{}".format(arch)
        try:
            path = Path(res).resolve()
            path.chmod(0o777)
            return path
        except NotImplementedError:
            logger.debug("System lacks support for chmod")
        except FileNotFoundError:
            logger.debug('Could not locate cos-tool at: "{}"'.format(res))
        return None

    def _exec(self, cmd: List[str]) -> str:
        result = subprocess.run(cmd, check=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        return result.stdout.decode("utf-8").strip()



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

import json
import logging
from typing import List
from unittest.mock import _Call, patch

import pytest

from src.cosl.loki_logger import LokiHandler


def _get_loki_http_post_payload(call: _Call):
    """Extract the POSTed json payload from the intercepted LokiLogger calls."""
    return json.loads(call.args[1].decode("utf-8"))


def _get_logs_severity(calls: List[_Call]):
    """Extract the loglevel from the intercepted LokiLogger calls."""
    return [
        _get_loki_http_post_payload(call)["streams"][0]["stream"]["severity"] for call in calls
    ]


def _get_loki_urls(calls: List[_Call]):
    """Extract the loki urls from the intercepted lokilogger calls."""
    return [call.args[0].full_url for call in calls]


@pytest.mark.parametrize("n_lokis", (1, 2, 5))
@pytest.mark.parametrize("loglevel_info", (True, False))
@patch("src.cosl.loki_logger.LokiEmitter._send_request")
def test_root_logging(send_request, n_lokis, loglevel_info):
    root_logger = logging.getLogger()

    urls = []
    created_handlers = []

    for i in range(n_lokis):
        url = f"http://loki_{i}.com"
        urls.append(url)
        handler = LokiHandler(url=url, labels={"test-label": f"loki-id-{i}"})

        root_logger.addHandler(handler)
        created_handlers.append(handler)

    try:
        any_logger = logging.getLogger("test-foo")
        if loglevel_info:
            any_logger.setLevel("INFO")

        any_logger.info("something")
        any_logger.error("something else")
    finally:
        # Cleanup the test env:
        #   We're mutating the root logger by adding handlers to it,
        #   that means any tests that will run after this one will also log to our
        #   handlers unless we do this cleanup.
        for handler in created_handlers:
            root_logger.removeHandler(handler)

    # check the intercepted calls.
    # we expect to see the info call and the error call, for every loki handler.
    # each info log gets sent to each loki handler
    # then each error log gets sent to each loki handler

    assert send_request.call_count == 2 * n_lokis

    assert _get_loki_urls(send_request.call_args_list) == urls * 2

    assert (
        _get_logs_severity(send_request.call_args_list) == ["info"] * n_lokis + ["error"] * n_lokis
    )



# Copyright 2020 Canonical Ltd.
# See LICENSE file for licensing details.

import json
import re
import unittest
import uuid
from pathlib import Path

import yaml
from deepdiff import DeepDiff
from fs.tempfs import TempFS

from cosl.juju_topology import JujuTopology
from cosl.rules import AlertRules


class TestEndpointProvider(unittest.TestCase):
    def test_each_alert_rule_is_topology_labeled(self):
        ri = AlertRules(
            query_type="promql",
            topology=JujuTopology(
                model="unittest",
                model_uuid=str(uuid.uuid4()),
                unit="tester/0",
                application="tester",
            ),
        )
        ri.add_path(Path(__file__).resolve().parent / "promql_rules" / "prometheus_alert_rules")

        alerts = ri.as_dict()
        self.assertIn("groups", alerts)
        self.assertEqual(len(alerts["groups"]), 5)
        for group in alerts["groups"]:
            for rule in group["rules"]:
                if "and_unit" not in group["name"]:
                    self.assertIn("labels", rule)
                    labels = rule["labels"]
                    self.assertIn("juju_model", labels)
                    self.assertIn("juju_application", labels)
                    self.assertIn("juju_model_uuid", labels)
                    # alerts should not have unit information if not already present
                    self.assertNotIn("juju_unit", rule["labels"])
                    self.assertNotIn("juju_unit=", rule["expr"])
                else:
                    self.assertIn("labels", rule)
                    labels = rule["labels"]
                    self.assertIn("juju_model", labels)
                    self.assertIn("juju_application", labels)
                    self.assertIn("juju_model_uuid", labels)
                    # unit information is already present
                    self.assertIn("juju_unit", rule["labels"])
                    self.assertIn("juju_unit=", rule["expr"])

    def test_each_alert_expression_is_topology_labeled(self):
        ri = AlertRules(
            query_type="promql",
            topology=JujuTopology(
                model="unittest",
                model_uuid=str(uuid.uuid4()),
                unit="tester/0",
                application="tester",
            ),
        )
        ri.add_path(Path(__file__).resolve().parent / "promql_rules" / "prometheus_alert_rules")

        alerts = ri.as_dict()
        self.assertIn("groups", alerts)
        self.assertEqual(len(alerts["groups"]), 5)
        group = alerts["groups"][0]
        for rule in group["rules"]:
            self.assertIn("expr", rule)
            for labels in expression_labels(rule["expr"]):
                self.assertIn("juju_model", labels)
                self.assertIn("juju_model_uuid", labels)
                self.assertIn("juju_application", labels)


def sorted_matchers(matchers) -> str:
    parts = [m.strip() for m in matchers.split(",")]
    return ",".join(sorted(parts))


def expression_labels(expr):
    """Extract labels from an alert rule expression.

    Args:
        expr: a string representing an alert expression.

    Returns:
        a generator which yields each set of labels in
        in the expression.
    """
    pattern = re.compile(r"\{.*\}")
    matches = pattern.findall(expr)
    for match in matches:
        match = match.replace("=", '":').replace("juju_", '"juju_')
        labels = json.loads(match)
        yield labels


class TestAlertRulesWithOneRulePerFile(unittest.TestCase):
    def setUp(self) -> None:
        free_standing_rule = {
            "alert": "free_standing",
            "expr": "avg(some_vector[5m]) > 5",
        }

        alert_rule = {
            "alert": "CPUOverUse",
            "expr": "process_cpu_seconds_total{%%juju_topology%%} > 0.12",
        }
        rules_file_dict = {"groups": [{"name": "group1", "rules": [alert_rule]}]}

        self.sandbox = TempFS("rule_files", auto_clean=True)
        self.addCleanup(self.sandbox.close)
        self.sandbox.makedirs("rules/prom/mixed_format")
        self.sandbox.makedirs("rules/prom/lma_format")
        self.sandbox.makedirs("rules/prom/prom_format")
        self.sandbox.writetext("rules/prom/mixed_format/lma_rule.rule", yaml.safe_dump(alert_rule))
        self.sandbox.writetext(
            "rules/prom/mixed_format/standard_rule.rule", yaml.safe_dump(rules_file_dict)
        )
        self.sandbox.writetext(
            "rules/prom/lma_format/free_standing_rule.rule", yaml.safe_dump(free_standing_rule)
        )
        self.sandbox.writetext(
            "rules/prom/prom_format/standard_rule.rule", yaml.safe_dump(rules_file_dict)
        )

        self.topology = JujuTopology(
            "MyModel", "12de4fae-06cc-4ceb-9089-567be09fec78", "MyApp", "MyUnit", "MyCharm"
        )

    def test_non_recursive_is_default(self):
        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(self.sandbox.getsyspath("/rules/prom/"))
        rules_file_dict = rules.as_dict()
        self.assertEqual({}, rules_file_dict)

    def test_non_recursive_lma_format_loading_from_root_dir(self):
        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(self.sandbox.getsyspath("/rules/prom/lma_format/"))
        rules_file_dict = rules.as_dict()

        expected_freestanding_rule = {
            "alert": "free_standing",
            "expr": "avg(some_vector[5m]) > 5",
            "labels": self.topology.label_matcher_dict,
        }

        expected_rules_file = {
            "groups": [
                {
                    "name": f"{sorted_matchers(self.topology.identifier)}_free_standing_rule_alerts",
                    "rules": [expected_freestanding_rule],
                },
            ]
        }

        self.assertEqual(expected_rules_file, rules_file_dict)

    def test_non_recursive_official_format_loading_from_root_dir(self):
        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(self.sandbox.getsyspath("/rules/prom/prom_format"))
        rules_file_dict = rules.as_dict()

        expected_alert_rule = {
            "alert": "CPUOverUse",
            "expr": f"process_cpu_seconds_total{{{sorted_matchers(self.topology.alert_expression_str)}}} > 0.12",
            "labels": self.topology.label_matcher_dict,
        }

        expected_rules_file = {
            "groups": [
                {
                    "name": f"{self.topology.identifier}_group1_alerts",
                    "rules": [expected_alert_rule],
                },
            ]
        }
        self.assertEqual(expected_rules_file, rules_file_dict)

    def test_alerts_in_both_formats_are_recursively_aggregated(self):
        """This test covers several aspects of the rules format.

        - Group name:
          - For rules in lma format, core group name is the filename
          - For rules in official format, core group name is the group name in the file
        """
        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(self.sandbox.getsyspath("/rules/prom"), recursive=True)
        rules_file_dict = rules.as_dict()

        expected_alert_rule = {
            "alert": "CPUOverUse",
            "expr": f"process_cpu_seconds_total{{{sorted_matchers(self.topology.alert_expression_str)}}} > 0.12",
            "labels": self.topology.label_matcher_dict,
        }

        expected_freestanding_rule = {
            "alert": "free_standing",
            "expr": "avg(some_vector[5m]) > 5",
            "labels": self.topology.label_matcher_dict,
        }

        expected_rules_file = {
            "groups": [
                {
                    "name": f"{self.topology.identifier}_mixed_format_group1_alerts",
                    "rules": [expected_alert_rule],
                },
                {
                    "name": f"{self.topology.identifier}_mixed_format_lma_rule_alerts",
                    "rules": [expected_alert_rule],
                },
                {
                    "name": f"{self.topology.identifier}_lma_format_free_standing_rule_alerts",
                    "rules": [expected_freestanding_rule],
                },
                {
                    "name": f"{self.topology.identifier}_prom_format_group1_alerts",
                    "rules": [expected_alert_rule],
                },
            ]
        }

        self.assertEqual({}, DeepDiff(expected_rules_file, rules_file_dict, ignore_order=True))

    def test_unit_not_in_alert_labels(self):
        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(self.sandbox.getsyspath("/rules/prom"), recursive=True)
        rules_file_dict = rules.as_dict()
        for group in rules_file_dict["groups"]:
            for rule in group["rules"]:
                self.assertTrue("juju_unit" not in rule["labels"])

    def test_charm_not_in_alert_expression(self):
        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(self.sandbox.getsyspath("/rules/prom"), recursive=True)
        rules_file_dict = rules.as_dict()
        for group in rules_file_dict["groups"]:
            for rule in group["rules"]:
                self.assertTrue("charm=" not in rule["expr"])


class TestAlertRulesWithMultipleRulesPerFile(unittest.TestCase):
    def setUp(self) -> None:
        self.topology = JujuTopology(
            "MyModel", "12de4fae-06cc-4ceb-9089-567be09fec78", "MyApp", "MyCharm"
        )

    def gen_rule(self, name, **extra):
        return {
            "alert": f"CPUOverUse_{name}",
            "expr": f"process_cpu_seconds_total{{{sorted_matchers(self.topology.label_matchers)}}} > 0.12",
            **extra,
        }

    def gen_group(self, name):
        return {"name": f"group_{name}", "rules": [self.gen_rule(1), self.gen_rule(2)]}

    def test_load_multiple_rules_per_file(self):
        """Test official format with multiple alert rules per group in multiple groups."""
        rules_file_dict = {"groups": [self.gen_group(1), self.gen_group(2)]}
        sandbox = TempFS("rule_files", auto_clean=True)
        sandbox.makedirs("rules")
        sandbox.writetext("rules/file.rule", yaml.safe_dump(rules_file_dict))

        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(sandbox.getsyspath("/rules"), recursive=False)
        rules_file_dict_read = rules.as_dict()

        expected_rules_file = {
            "groups": [
                {
                    "name": f"{self.topology.identifier}_group_1_alerts",
                    "rules": [
                        self.gen_rule(1, labels=self.topology.label_matcher_dict),
                        self.gen_rule(2, labels=self.topology.label_matcher_dict),
                    ],
                },
                {
                    "name": f"{self.topology.identifier}_group_2_alerts",
                    "rules": [
                        self.gen_rule(1, labels=self.topology.label_matcher_dict),
                        self.gen_rule(2, labels=self.topology.label_matcher_dict),
                    ],
                },
            ]
        }
        self.assertDictEqual(expected_rules_file, rules_file_dict_read)

    def test_duplicated_alert_names_within_alert_rules_list_are_silently_accepted(self):
        """Test official format when the alert rules list has a duplicated alert name."""
        rules_file_dict = {
            "groups": [
                {
                    "name": "my_group",
                    "rules": [self.gen_rule("same"), self.gen_rule("same")],
                }
            ]
        }
        sandbox = TempFS("rule_files", auto_clean=True)
        sandbox.makedirs("rules")
        sandbox.writetext("rules/file.rule", yaml.safe_dump(rules_file_dict))

        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(sandbox.getsyspath("/rules"), recursive=False)
        rules_file_dict_read = rules.as_dict()

        expected_rules_file = {
            "groups": [
                {
                    "name": f"{self.topology.identifier}_my_group_alerts",
                    "rules": [
                        self.gen_rule("same", labels=self.topology.label_matcher_dict),
                        self.gen_rule("same", labels=self.topology.label_matcher_dict),
                    ],
                },
            ]
        }
        self.assertDictEqual(expected_rules_file, rules_file_dict_read)

    def test_duplicated_group_names_within_a_file_are_silently_accepted(self):
        rules_file_dict = {"groups": [self.gen_group("same"), self.gen_group("same")]}
        sandbox = TempFS("rule_files", auto_clean=True)
        sandbox.makedirs("rules")
        sandbox.writetext("rules/file.rule", yaml.safe_dump(rules_file_dict))

        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(sandbox.getsyspath("/rules"), recursive=False)
        rules_file_dict_read = rules.as_dict()

        expected_rules_file = {
            "groups": [
                {
                    "name": f"{self.topology.identifier}_group_same_alerts",
                    "rules": [
                        self.gen_rule(1, labels=self.topology.label_matcher_dict),
                        self.gen_rule(2, labels=self.topology.label_matcher_dict),
                    ],
                },
                {
                    "name": f"{self.topology.identifier}_group_same_alerts",
                    "rules": [
                        self.gen_rule(1, labels=self.topology.label_matcher_dict),
                        self.gen_rule(2, labels=self.topology.label_matcher_dict),
                    ],
                },
            ]
        }
        self.assertDictEqual(expected_rules_file, rules_file_dict_read)

    def test_deeply_nested(self):
        sandbox = TempFS("rule_files", auto_clean=True)
        sandbox.makedirs("rules/a/b/")
        sandbox.writetext("rules/file.rule", yaml.safe_dump(self.gen_rule(0)))
        sandbox.writetext("rules/a/file.rule", yaml.safe_dump(self.gen_rule(1)))
        sandbox.writetext("rules/a/b/file.rule", yaml.safe_dump(self.gen_rule(2)))

        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(sandbox.getsyspath("/rules"), recursive=True)
        rules_file_dict_read = rules.as_dict()

        expected_rules_file = {
            "groups": [
                {
                    "name": f"{self.topology.identifier}_file_alerts",
                    "rules": [self.gen_rule(0, labels=self.topology.label_matcher_dict)],
                },
                {
                    "name": f"{self.topology.identifier}_a_file_alerts",
                    "rules": [self.gen_rule(1, labels=self.topology.label_matcher_dict)],
                },
                {
                    "name": f"{self.topology.identifier}_a_b_file_alerts",
                    "rules": [self.gen_rule(2, labels=self.topology.label_matcher_dict)],
                },
            ]
        }
        self.assertDictEqual(expected_rules_file, rules_file_dict_read)


class TestAlertRulesContainingUnitTopology(unittest.TestCase):
    """Tests that check we does not remove unit topology.

    Unit Topology information is not added to alert rules expressions and labels,
    by the Provider. However, if unit topology information is
    present in the labels then it must not be removed since the client that
    the alert be limited to a specific unit.
    """

    def test_unit_label_is_retained_if_hard_coded(self):
        ri = AlertRules(
            query_type="promql",
            topology=JujuTopology(
                model="unittest",
                model_uuid=str(uuid.uuid4()),
                unit="tester/0",
                application="tester",
            ),
        )
        ri.add_path(
            Path(__file__).resolve().parent / "promql_rules" / "alert_rules_with_unit_topology"
        )

        alert_rules = ri.as_dict()
        for group in alert_rules["groups"]:
            for rule in group["rules"]:
                self.assertIn("juju_unit", rule["labels"])
                self.assertIn("juju_unit=", rule["expr"])



import json
from pathlib import Path
from unittest.mock import MagicMock, patch

import ops
import pytest
import tenacity
import yaml
from ops import Framework
from ops.pebble import Layer, ServiceStatus
from scenario import Container, Context, Mount, Relation, State
from scenario.runtime import UncaughtCharmError

from cosl.coordinated_workers.worker import CONFIG_FILE, Worker


@pytest.fixture(autouse=True)
def patch_running_version():
    with patch("cosl.coordinated_workers.worker.Worker.running_version", new=lambda _: "42.42"):
        yield


class MyCharm(ops.CharmBase):
    layer = Layer("")

    def __init__(self, framework: Framework):
        super().__init__(framework)
        self.worker = Worker(self, "foo", lambda _: self.layer, {"cluster": "cluster"})


def test_no_roles_error():
    # Test that a charm that defines NO 'role-x' config options, when run,
    # raises a WorkerError

    # WHEN you define a charm with no role-x config options
    ctx = Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={},
    )

    # IF the charm executes any event
    # THEN the charm raises an error
    with pytest.raises(UncaughtCharmError):
        ctx.run("update-status", State(containers=[Container("foo")]))


@pytest.mark.parametrize(
    "roles_active, roles_inactive, expected",
    (
        (
            ["read", "write", "ingester", "all"],
            ["alertmanager"],
            ["read", "write", "ingester", "all"],
        ),
        (["read", "write"], ["alertmanager"], ["read", "write"]),
        (["read"], ["alertmanager", "write", "ingester", "all"], ["read"]),
        ([], ["read", "write", "ingester", "all", "alertmanager"], []),
    ),
)
def test_roles_from_config(roles_active, roles_inactive, expected):
    # Test that a charm that defines any 'role-x' config options, when run,
    # correctly determines which ones are enabled through the Worker

    # WHEN you define a charm with a few role-x config options
    ctx = Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={
            "options": {
                f"role-{r}": {"type": "boolean", "default": "false"}
                for r in (roles_active + roles_inactive)
            }
        },
    )

    # AND the charm runs with a few of those set to true, the rest to false
    with ctx.manager(
        "update-status",
        State(
            containers=[Container("foo")],
            config={
                **{f"role-{r}": False for r in roles_inactive},
                **{f"role-{r}": True for r in roles_active},
            },
        ),
    ) as mgr:
        # THEN the Worker.roles method correctly returns the list of only those that are set to true
        assert set(mgr.charm.worker.roles) == set(expected)


def test_worker_restarts_if_some_service_not_up(tmp_path):
    # GIVEN a worker with some services
    MyCharm.layer = Layer(
        {
            "services": {
                "foo": {
                    "summary": "foos all the things",
                    "description": "bar",
                    "startup": "enabled",
                    "override": "merge",
                    "command": "ls -la",
                },
                "bar": {
                    "summary": "bars the foos",
                    "description": "bar",
                    "startup": "enabled",
                    "command": "exit 1",
                },
                "baz": {
                    "summary": "bazzes all of the bars",
                    "description": "bar",
                    "startup": "enabled",
                    "command": "echo hi",
                },
            }
        }
    )
    ctx = Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    # WHEN the charm receives any event and there are no changes to the config or the layer,
    #  but some of the services are down
    cfg = tmp_path / "cfg.yaml"
    cfg.write_text("some: yaml")
    container = Container(
        "foo",
        can_connect=True,
        mounts={"local": Mount(CONFIG_FILE, cfg)},
        service_status={
            "foo": ServiceStatus.INACTIVE,
            "bar": ServiceStatus.ACTIVE,
            "baz": ServiceStatus.INACTIVE,
        },
    )
    state_out = ctx.run(container.pebble_ready_event, State(containers=[container]))

    # THEN the charm restarts all the services that are down
    container_out = state_out.get_container("foo")
    service_statuses = container_out.service_status.values()
    assert all(svc is ServiceStatus.ACTIVE for svc in service_statuses), [
        stat.value for stat in service_statuses
    ]


def test_worker_does_not_restart_external_services(tmp_path):
    # GIVEN a worker with some services and a layer with some other services
    MyCharm.layer = Layer(
        {
            "services": {
                "foo": {
                    "summary": "foos all the things",
                    "override": "merge",
                    "description": "bar",
                    "startup": "enabled",
                    "command": "ls -la",
                }
            }
        }
    )
    other_layer = Layer(
        {
            "services": {
                "bar": {
                    "summary": "bars the foos",
                    "description": "bar",
                    "startup": "enabled",
                    "command": "exit 1",
                },
                "baz": {
                    "summary": "bazzes all of the bars",
                    "description": "bar",
                    "startup": "enabled",
                    "command": "echo hi",
                },
            }
        }
    )

    ctx = Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    # WHEN the charm receives any event and there are no changes to the config or the layer,
    #  but some of the services are down
    cfg = tmp_path / "cfg.yaml"
    cfg.write_text("some: yaml")
    container = Container(
        "foo",
        can_connect=True,
        mounts={"local": Mount(CONFIG_FILE, cfg)},
        layers={"foo": MyCharm.layer, "bar": other_layer},
        service_status={
            # layer foo has some inactive
            "foo": ServiceStatus.INACTIVE,
            # layer bar has some inactive
            "bar": ServiceStatus.ACTIVE,
            "baz": ServiceStatus.INACTIVE,
        },
    )
    state_out = ctx.run(container.pebble_ready_event, State(containers=[container]))

    # THEN the charm restarts all the services that are down
    container_out = state_out.get_container("foo")
    assert container_out.service_status == {
        # layer foo service is now active
        "foo": ServiceStatus.ACTIVE,
        # layer bar services is unchanged
        "bar": ServiceStatus.ACTIVE,
        "baz": ServiceStatus.INACTIVE,
    }


def test_worker_raises_if_service_restart_fails_for_too_long(tmp_path):
    # GIVEN a worker with some services
    MyCharm.layer = Layer(
        {
            "services": {
                "foo": {
                    "summary": "foos all the things",
                    "description": "bar",
                    "startup": "enabled",
                    "command": "ls -la",
                },
            }
        }
    )
    ctx = Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    cfg = tmp_path / "cfg.yaml"
    cfg.write_text("some: yaml")
    container = Container(
        "foo",
        can_connect=True,
        mounts={"local": Mount(CONFIG_FILE, cfg)},
        service_status={
            "foo": ServiceStatus.INACTIVE,
        },
    )

    # WHEN service restart fails
    def raise_change_error(*args):
        raise ops.pebble.ChangeError("something", MagicMock())

    with patch("ops.model.Container.restart", new=raise_change_error):
        # THEN the charm errors out
        with pytest.raises(Exception):
            # technically an ops.pebble.ChangeError but the context manager doesn't catch it for some reason

            with ctx.manager(container.pebble_ready_event, State(containers=[container])) as mgr:
                # so we don't have to wait for minutes:
                mgr.charm.worker.SERVICE_START_RETRY_WAIT = tenacity.wait_none()
                mgr.charm.worker.SERVICE_START_RETRY_STOP = tenacity.stop_after_delay(2)


@pytest.mark.parametrize(
    "remote_databag, expected",
    (
        (
            {
                "remote_write_endpoints": json.dumps([{"url": "test-url.com"}]),
                "worker_config": json.dumps("test"),
            },
            [{"url": "test-url.com"}],
        ),
        ({"remote_write_endpoints": json.dumps(None), "worker_config": json.dumps("test")}, []),
        (
            {
                "remote_write_endpoints": json.dumps(
                    [{"url": "test-url.com"}, {"url": "test2-url.com"}]
                ),
                "worker_config": json.dumps("test"),
            },
            [{"url": "test-url.com"}, {"url": "test2-url.com"}],
        ),
    ),
)
def test_get_remote_write_endpoints(remote_databag, expected):
    ctx = Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    container = Container(
        "foo",
        can_connect=True,
    )
    relation = Relation(
        "cluster",
        remote_app_data=remote_databag,
    )
    with ctx.manager(
        relation.changed_event, State(containers=[container], relations=[relation])
    ) as mgr:
        charm = mgr.charm
        mgr.run()
        assert charm.worker.cluster.get_remote_write_endpoints() == expected


def test_config_preprocessor():
    # GIVEN a charm with a config preprocessor
    new_config = {"modified": "config"}

    class MyWorker(Worker):

        @property
        def _worker_config(self):
            # mock config processor that entirely replaces the config with another,
            # normally one would call super and manipulate
            return new_config

    class MyCharm(ops.CharmBase):
        layer = Layer({"services": {"foo": {"command": ["bar"]}}})

        def __init__(self, framework: Framework):
            super().__init__(framework)
            self.worker = MyWorker(
                self,
                "foo",
                lambda _: self.layer,
                {"cluster": "cluster"},
            )

    ctx = Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={
            "options": {
                "role-all": {"type": "boolean", "default": "true"},
                "role-none": {"type": "boolean", "default": "false"},
            }
        },
    )

    # WHEN the charm writes the config to disk
    state_out = ctx.run(
        "config_changed",
        State(
            config={"role-all": True},
            containers=[Container("foo", can_connect=True)],
            relations=[
                Relation(
                    "cluster",
                    remote_app_data={
                        "worker_config": json.dumps(yaml.safe_dump({"original": "config"}))
                    },
                )
            ],
        ),
    )

    # THEN the data gets preprocessed
    fs = Path(str(state_out.get_container("foo").get_filesystem(ctx)) + CONFIG_FILE)
    assert fs.read_text() == yaml.safe_dump(new_config)



# Copyright 2024 Canonical Ltd.
# See LICENSE file for licensing details.

"""Tooling for rendering charm status message for mandatory relation pairs."""

from typing import Dict, List, Set


class MandatoryRelationPairs:
    """A helper class for ensuring that an incoming relation has a matching outgoing relation.

    >>> # "rel1" must be paired with (both "r1" and "r2") or "r3", and "rel2" with "r4"
    >>> mrp = MandatoryRelationPairs({"rel1": [{"r1", "r2"}, {"r3"}], "rel2": [{"r4"}]})
    >>> mrp.get_missing_as_str("rel1", "r1", "rel2")
    "['r2']|['r3'] for rel1; ['r4'] for rel2"
    >>> mrp.get_missing_as_str("rel1", "r3")
    ''
    >>> mrp.get_missing_as_str()
    ''
    """

    def __init__(self, pairs: Dict[str, List[Set[str]]]):
        # pairs look like this:
        # {
        #     "cos-agent": [  # must be paired with:
        #         {"grafana-cloud-config"},  # or
        #         {"send-remote-write", "logging-consumer", "grafana-dashboards-provider"},
        #     ],
        #     "juju-info": [  # must be paired with:
        #         {"grafana-cloud-config"},  # or
        #         {"send-remote-write", "logging-consumer"},
        #     ],
        # }
        self._pairs = pairs

    def get_missing(self, *relations_present: str) -> Dict[str, List[Set[str]]]:
        """Returns a mapping from relation name to the set of missing mandatory relation names.

        If nothing is missing for a given relation, then it won't be listed (as a key) in the dict.
        """
        # From all relations currently present, get a set of 'incoming' relations that must
        # have matching 'outgoing' relations.
        relations_incoming: Set[str] = set(relations_present).intersection(self._pairs.keys())

        # Same shape as `self._pairs`
        missing = {
            rel: [mandatory.difference(relations_present) for mandatory in self._pairs[rel]]
            for rel in relations_incoming
        }

        # If any rel in missing has an empty set, it means that it has at least one combo of
        # matching relations, so that rel can be dropped out of `missing`.
        missing = {k: v for k, v in missing.items() if all(v)}

        return missing

    def get_missing_as_str(self, *relations_present: str) -> str:
        """Return the missing relations, formatted into a string."""
        missing = self.get_missing(*relations_present)

        combos = "; ".join(
            sorted(
                [
                    "|".join(sorted([str(sorted(mandatory)) for mandatory in v])) + f" for {rel}"
                    for rel, v in sorted(missing.items())
                ]
            )
        )

        return f"{combos}" if combos else ""



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

"""Grafana Dashboard."""

import base64
import json
import logging
import lzma
from typing import Any, Dict, Union

logger = logging.getLogger(__name__)


class GrafanaDashboard(str):
    """GrafanaDashboard represents an actual dashboard in Grafana.

    The class is used to compress and encode, or decompress and decode,
    Grafana Dashboards in JSON format using LZMA.
    """

    @staticmethod
    def _serialize(raw_json: Union[str, bytes]) -> "GrafanaDashboard":
        if not isinstance(raw_json, bytes):
            raw_json = raw_json.encode("utf-8")
        encoded = base64.b64encode(lzma.compress(raw_json)).decode("utf-8")
        return GrafanaDashboard(encoded)

    def _deserialize(self) -> Dict[str, Any]:
        try:
            raw = lzma.decompress(base64.b64decode(self.encode("utf-8"))).decode()
            return json.loads(raw)
        except json.decoder.JSONDecodeError as e:
            logger.error("Invalid Dashboard format: %s", e)
            return {}

    def __repr__(self):
        """Return string representation of self."""
        return "<GrafanaDashboard>"



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

import json
import unittest

from cosl import GrafanaDashboard


class TestDashboard(unittest.TestCase):
    """Tests the GrafanaDashboard class."""

    def test_serializes_and_deserializes(self):
        expected_output = {"msg": "this is the expected output after passing through the class."}

        dash = GrafanaDashboard._serialize(json.dumps(expected_output))

        assert dash._deserialize() == expected_output
